/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

/* i/o layering support */

#ifndef AP_IOL_H
#define AP_IOL_H

#include "apr_general.h" /* For ap_s?size_t */
#include "apr_errno.h" /* For ap_status_t and the APR_errnos */

typedef struct ap_iol ap_iol;
typedef struct ap_iol_methods ap_iol_methods;

typedef enum {
    AP_IOL_TIMEOUT
} ap_iol_option;

/*
    write, writev, read guarantee they won't return APR_EINTR.

    If no error occurs, they return APR_SUCCESS.

    If timeout is negative, they may block forever, and they still don't
    guarantee they'll write/read the full length.

    If timeout is 0, they will never block, and will return APR_EAGAIN
    in the event that they can't read/write any bytes.

    If the timeout is positive, they will block up to the specified number
    of seconds.  If the read/write can't be completed in that time,
    APR_ETIMEDOUT will be returned.

    Regardless of the return value, bytes may have been read or written.
    (For example, APR_EOF may be returned along with 100 bytes from read.)

    XXX: timeouts should probably be redefined to be in microseconds to match APR.

    Just to drive the point home again -- none of these functions guarantee
    they will read/write the entire length specified.

    Note: timeout == 0 is non-blocking operation... which requires a lot
    of care if you're implementing filters.  It may mean you have to keep
    state from one call to the next.  The chunking filter will have an
    example of this.  If you're not sure how to handle the state sitatuation
    you will want to return APR_EINVAL when the timeout is set to 0.  This may
    break various things... hopefully we'll be able to recover gracefully.

    Note: implementation tip: if your system doesn't have writev(), you
    still have to implement a writev() method.  However, since the
    iol writev() is allowed to return a partial write, you can essentially
    treat all writev()s as write(fd, vec[0].iov_base, vec[0].iov_len),
    and the layers above will loop over the entire vector for you.
*/

struct ap_iol_methods {
    ap_status_t (*close)(ap_iol *fd);
    ap_status_t (*write)(ap_iol *fd, const char *buf, ap_size_t len,
                         ap_ssize_t *nbytes);
    ap_status_t (*writev)(ap_iol *fd, const struct iovec *vec, ap_size_t nvec,
                          ap_ssize_t *nbytes);
    ap_status_t (*read)(ap_iol *fd, char *buf, ap_size_t len,
                        ap_ssize_t *nbytes);
    ap_status_t (*setopt)(ap_iol *fd, ap_iol_option opt, const void *value);
    ap_status_t (*getopt)(ap_iol *fd, ap_iol_option opt, void *value);
    ap_status_t (* sendfile)(ap_iol *fd, ap_file_t * file, ap_hdtr_t * hdtr, 
                             ap_off_t * offset, ap_size_t * len, 
                             ap_int32_t flags);
    ap_status_t (*shutdown)(ap_iol *fd, int how);
    /* TODO: accept, connect, ... */
};

/* it is expected that each io layer will extend this structure as they
    require; they must only ensure that the methods pointer remains in
    the indicated position. */
struct ap_iol {
    const ap_iol_methods *methods;
};

#define iol_close(iol)	((iol)->methods->close((iol)))
#define iol_write(iol, a, b, c) ((iol)->methods->write((iol), (a), (b), (c)))
#define iol_writev(iol, a, b, c) ((iol)->methods->writev((iol), (a), (b), (c)))
#define iol_read(iol, a, b, c) ((iol)->methods->read((iol), (a), (b), (c)))
#define iol_setopt(iol, a, b) ((iol)->methods->setopt((iol), (a), (b)))
#define iol_getopt(iol, a, b) ((iol)->methods->getopt((iol), (a), (b)))
#define iol_sendfile(iol, a, b, c, d, e) ((iol)->methods->sendfile((iol), (a), (b), (c), (d), (e)))
#define iol_shutdown(iol, a) ((iol)->methods->shutdown((iol), (a)))

/* the file iol */
ap_iol *ap_create_file_iol(ap_file_t *file);

#endif
